//----------------------------------------------------------------------------
//
// File:        disk-serializer-pc99.cpp
// Date:        06-July-2015
// Programmer:  Marc Rousseau
//
// Description: A class to load/save PC99 disk images
//
// Copyright (c) 2015 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <cstring>
#include "common.hpp"
#include "logger.hpp"
#include "disk-media.hpp"
#include "disk-serializer-pc99.hpp"
#include "idisk-track.hpp"

DBG_REGISTER( __FILE__ );

static void ZapCRC( sDataBuffer &buffer, std::vector<int> clocks, track::Format format )
{
	int offset = ( format == track::Format::MFM ) ? 1 : 0;

	int lastSize = 0;

	// Revert CRCs back to F7F7
	for( auto clock : clocks )
	{
		if(( size_t ) ( clock + offset ) < buffer.size( ))
		{
			UINT8 *ptr = &buffer[ clock + offset ];
			switch( *ptr )
			{
				case MARK_IDAM :
					if(( size_t ) ( clock + offset + 4 ) < buffer.size( ))
					{
						lastSize = 128 << ( ptr[ 4 ] & 0x03 );
						ptr[ 5 ] = 0xF7;
						ptr[ 6 ] = 0xF7;
					}
					break;
				case MARK_DDAM :
				case MARK_DAMx :
				case MARK_DAMy :
				case MARK_DAM :
					if(( size_t ) ( clock + offset + lastSize + 2 ) < buffer.size( ))
					{
						ptr[ lastSize + 1 ] = 0xF7;
						ptr[ lastSize + 2 ] = 0xF7;
					}
					break;
				default:
					break;
			}
		}
	}
}

cDiskSerializerPC99::cDiskSerializerPC99( ) :
	cBaseObject( "cDiskSerializerPC99" )
{
}

cDiskSerializerPC99::~cDiskSerializerPC99( )
{
}

bool cDiskSerializerPC99::MatchesFormat( FILE *file )
{
	FUNCTION_ENTRY( nullptr, "cDiskSerializerPC99::MatchesFormat", true );

	// TODO

	return true;
}

bool cDiskSerializerPC99::SupportsFeatures( const cDiskImage &image )
{
	FUNCTION_ENTRY( this, "cDiskSerializerPC99::SupportsFeatures", true );

	// TODO

	return true;
}

eDiskFormat cDiskSerializerPC99::GetFormat( ) const
{
	FUNCTION_ENTRY( this, "cDiskSerializerPC99::GetFormat", true );

	return FORMAT_RAW_TRACK;
}

//----------------------------------------------------------------------------
//
// Read disk files that contain raw track data.  This is the format used by
// PC99.  This code attempts to correctly read disk images that may contain
// unusual formatting.
//
//----------------------------------------------------------------------------

bool cDiskSerializerPC99::ReadFile( FILE *file, cDiskImage *image )
{
	FUNCTION_ENTRY( this, "cDiskSerializerPC99::ReadFile", true );

	UINT8 buffer[ 64 ];

	// Find out the index GAP character (should be 0xFF or 0x4E but PC99 uses 0x00 for FM disks)
	if( fread( buffer, sizeof( buffer ), 1, file ) != 1 )
	{
		DBG_ERROR( "Error reading from file" );
		return false;
	}

	track::Format format = ( buffer[ 0 ] != 0x4E ) ? track::Format::FM : track::Format::MFM;

	size_t maxTrackSize = 120 * (( format == track::Format::FM ) ? TRACK_SIZE_FM : TRACK_SIZE_MFM ) / 100;

	if( FindAddressMark( 0xFC, 0xFC, format, buffer, buffer + 64 ) == nullptr )
	{
		return false;
	}

	auto trackData = ReadTrackData( file, format, maxTrackSize );

	int maxTrack = MAX_TRACKS_HI;

	// Clear any old data & prepare for a new image
	image->AllocateTracks( MAX_TRACKS, 2 );

	// Try to distinguish between single-sided 80 tracks and double-sided 40 track disks
	if( trackData.size( ) == 80 )
	{
		auto track = image->GetTrack( 0, 0 );

		track->RawWrite( format, trackData[ 40 ].clock, trackData[ 40 ].data );

		auto sector = track->GetSector( -1, -1, -1 );

		if( sector->LogicalHead( ) == 1 )
		{
			image->AllocateTracks( MAX_TRACKS_LO, 2 );
			maxTrack = MAX_TRACKS_LO;
		}
	}

	size_t index = 0;
	for( int h = 0; h < 2; h++ )
	{
		for( int t = 0; t < maxTrack; t++ )
		{
			if(( index < trackData.size( )) && ( trackData[ index ].data.size( ) > 0 ))
			{
				auto track = dynamic_cast<cDiskTrack*>( image->GetTrack( t, h ));

				track->RawWrite( format, trackData[ index ].clock, trackData[ index ].data );

				// Update the CRC values read in (PC99 stores them as F7F7)
				for( auto &sector : track->GetSectors( ))
				{
					track->DataModified( sector->GetID( ) - 1, 4 );
					track->DataModified( sector->GetData( ) - 1, sector->Size( ));
				}

				index++;
			}
		}
	}

	DBG_EVENT( "Disk loaded" );

	return true;
}

bool cDiskSerializerPC99::WriteFile( const cDiskImage &image, FILE *file )
{
	FUNCTION_ENTRY( this, "cDiskSerializerPC99::WriteFile", true );

	for( int h = 0; h < image.GetNumHeads( ); h++ )
	{
		for( int t = 0; t < image.GetNumTracks( ); t++ )
		{
			auto track = image.GetTrack( t, h );

			auto buffer = track->Read( );

			if( !buffer.empty( ))
			{
				ZapCRC( buffer, track->GetClockLocations( ), track->GetFormat( ));

				if( fwrite( buffer.data( ), buffer.size( ), 1, file ) != 1 )
				{
					DBG_ERROR( "Error writing to file" );
					return false;
				}
			}
		}
	}

	return true;
}

const UINT8 *cDiskSerializerPC99::FindAddressMark( UINT8 mask, UINT8 mark, track::Format format, const UINT8 *ptr, const UINT8 *max )
{
	FUNCTION_ENTRY( nullptr, "cDiskSerializerPC99::FindAddressMark", false );

restart:

	// Find start of sync bytes
	while( *ptr != '\x00' )
	{
		if( ++ptr >= max )
		{
			return nullptr;
		}
	}

	while( *ptr == '\x00' )
	{
		if( ++ptr >= max )
		{
			return nullptr;
		}
	}

	// Look for clock pattern for MFM tracks
	if( format == track::Format::MFM )
	{
		for( int i = 0; i < 3; i++ )
		{
			if( *ptr != 0xA1 )
			{
				goto restart;
			}
			if( ++ptr >= max )
			{
				return nullptr;
			}
		}
	}

	if(( *ptr++ & mask ) != mark )
	{
		goto restart;
	}

	return ptr;
}

cDiskSerializerPC99::sTrackInfo cDiskSerializerPC99::FindTrack( track::Format format, const UINT8 *start, const UINT8 *max )
{
	FUNCTION_ENTRY( nullptr, "cDiskSerializerPC99::FindTrack", true );

	sTrackInfo track;

	const UINT8 *ptr = start;

	// The track should start with either an Index (0xFC) or ID (0xFE) Address Mark
	const UINT8 *markID = FindAddressMark( 0xFC, 0xFC, format, ptr, max );

	if( markID != nullptr )
	{
		// Remember the number of bytes from the beginning of the track to the 1st Address Mark
		int indexGapSize = markID - ptr - 1;

		// Handle an Index Address Mark
		if( markID[ -1 ] == MARK_IAM )
		{
			track.clock.push_back( markID - start - 1 );

			ptr = markID;
		}

		while( ptr < max )
		{
			markID = FindAddressMark( 0xFC, 0xFC, format, ptr, max );

			if( markID == nullptr )
			{
				break;
			}

			int gapSize = markID - ptr - 1;

			// See if we have passed the EOT filler
			if( gapSize > (( format == track::Format::FM ) ? 225 : 736 ))
			{
				ptr = markID - indexGapSize - 1;
				break;
			}

			if( markID[ -1 ] != MARK_IDAM )
			{
				break;
			}

			track.clock.push_back( markID - start - 1 );

			// Handle an ID Address Mark
			if( markID[ -1 ] == MARK_IDAM )
			{
				ptr = markID + 4 + 2;

				int dataSize = 128 << markID[ 3 ];

				// NOTE: WD1771 controllers can use 0xF8 0xF9 0xFA or 0xFB
				// NOTE: PC controllers cannot detect 0xF9 or 0xFA marks
				const UINT8 *markData = FindAddressMark( 0xFC, 0xF8, format, ptr, max );
				if( markData == nullptr )
				{
					break;
				}

				track.clock.push_back( markData - start - 1 );

				ptr = markData + dataSize + 2;
			}
		}
	}

	track.data = { start, ptr };

	return track;
}

std::vector<cDiskSerializerPC99::sTrackInfo> cDiskSerializerPC99::ReadTrackData( FILE *file, track::Format format, size_t maxTrackSize )
{
	FUNCTION_ENTRY( nullptr, "cDiskSerializerPC99::ReadTrackData", true );

	fseek( file, 0, SEEK_SET );

	std::vector<sTrackInfo> trackData;

	size_t count = 0;

	UINT8 trackBuffer[ 2 * MAX_TRACK_SIZE ];

	while( count < sizeof( trackBuffer ))
	{
		size_t read = ( int ) fread( trackBuffer + count, 1, sizeof( trackBuffer ) - count, file );

		if(( read == ( size_t ) EOF ) || ( read == 0 ))
		{
			break;
		}

		count += read;

		// trackBuffer should now contain enough data for at least 1 track
		while( count >= maxTrackSize )
		{
			auto track = FindTrack( format, trackBuffer, trackBuffer + count );
			if( track.data.empty( ))
			{
				break;
			}

			trackData.push_back( track );

			count -= track.data.size( );

			memmove( trackBuffer, trackBuffer + track.data.size( ), count );
		}
	}

	if( count > 0 )
	{
		auto track = FindTrack( format, trackBuffer, trackBuffer + count );
		if( !track.data.empty( ))
		{
			trackData.push_back( track );
		}
	}

	return trackData;
}
